#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <gccore.h>
#include <ogcsys.h>
#include <math.h>
#include <wiiuse/wpad.h>

#include "GRRLIB.h"
#include "font5.h"
#include "wbwiimote.h"

extern u16 *GRRLIB_buffer;

WBPoint calibPoints1[] = {
  { .1, .1 },
  { .9, .1 },
  { .1, .9 },
  { .9, .9 }
};
WBPoint calibPoints2[] = {
  { 0, 0 },
  { 0, 0 },
  { 0, 0 },
  { 0, 0 }
};

const int numCalibPoints = 4;
int numCalibPointsDone = 0;
const int calibPointRadius = 3;

u16 colors[] = {
  0x0000,
  0xf000,
  0x0ff0,
  0x000f,
  0xff00,
  0x00ff,
  0xf00f,
  0xfff0,
  0xff0f,
  0xf0ff,
  0x0fff,
  0xffff
};
const int numcolors = 12;
int colorindex = 0;

#define TOOL_LINE 0
#define TOOL_SMALLDOT 1
#define TOOL_LARGEDOT 2
const int numtools = 3;
int tool;

const WBPoint screenDim = { 640, 480 };
const WBPoint wiimoteDim = { 1024, 768 };
const WBPoint zero = {0, 0};
WBPoint held = {0, 0};
WBPoint cursor = {0, 0}, oldcursor = {0, 0}, startcursor = {0, 0};

WBPoint coldecrP1, coldecrP2, colincrP1, colincrP2;
WBPoint tooldecrP1, tooldecrP2, toolincrP1, toolincrP2;

WPADData* wpads;

char msg[255];

int wiimotenum = 0, wbwiimotenum = 1;
int needsfillscreen = 0;

typedef enum {
  STATE_TITLE,
  STATE_TEST,
  STATE_CALIBRATING,
  STATE_DRAWING
} State;

State state = STATE_TITLE;

int buttonPushed(buttons)
{
  return (WPAD_ButtonsDown(wiimotenum) & buttons ||
	  WPAD_ButtonsDown(wbwiimotenum) & buttons);
}

WBPoint normToScreen(WBPoint pt)
{
  WBPoint rv;
  rv.u = pt.u*screenDim.u;
  rv.v = pt.v*screenDim.v;
  return rv;
}

WBPoint screenToNorm(WBPoint pt)
{
  WBPoint rv;
  rv.u = pt.u/screenDim.u;
  rv.v = pt.v/screenDim.v;
  return rv;
}

WBPoint screenToNormXY(float x, float y)
{
  WBPoint rv;
  rv.u = x/screenDim.u;
  rv.v = y/screenDim.v;
  return rv;
}

WBPoint normToWiimote(WBPoint pt)
{
  WBPoint rv;
  rv.u = pt.u*wiimoteDim.u;
  rv.v = pt.v*wiimoteDim.v;
  return rv;
}

WBPoint wiimoteToNorm(WBPoint pt)
{
  WBPoint rv;
  rv.u = pt.u/wiimoteDim.u;
  rv.v = pt.v/wiimoteDim.v;
  return rv;
}

int isPointZero(WBPoint pt)
{
  return pt.u == 0 || pt.v == 0;
}

double distance(WBPoint pt1, WBPoint pt2)
{
  return sqrt((pt1.u - pt2.u)*(pt1.u - pt2.u) +
	      (pt1.v - pt2.v)*(pt1.v - pt2.v));
}

void reboot()
{
  VIDEO_WaitVSync();
  int fd = IOS_Open("/dev/stm/immediate", 0);
  IOS_Ioctl(fd, 0x2001, NULL, 0, NULL, 0);
  IOS_Close(fd);
}

int inbox(WBPoint pt, WBPoint boxp1, WBPoint boxp2)
{
  return (pt.u >= boxp1.u && pt.u <= boxp2.u &&
	  pt.v >= boxp1.v && pt.v <= boxp2.v);
}

void wbclick(WBPoint atstart, WBPoint atend)
{
  if(inbox(atstart, coldecrP1, coldecrP2) &&
     inbox(atend, coldecrP1, coldecrP2))
    {
      colorindex--;
      if(colorindex < 0) colorindex = numcolors-1;
    }
  else if(inbox(atstart, colincrP1, colincrP2) &&
	  inbox(atend, colincrP1, colincrP2))
    {
      colorindex++;
      if(colorindex >= numcolors) colorindex = 0;
    }

  else if(inbox(atstart, tooldecrP1, tooldecrP2) &&
	  inbox(atend, tooldecrP1, tooldecrP2))
    {
      tool--;
      if(tool < 0) tool = numtools-1;
    }
  else if(inbox(atstart, toolincrP1, toolincrP2) &&
	  inbox(atend, toolincrP1, toolincrP2))
    {
      tool++;
      if(tool >= numtools) tool = 0;
    }
}

void drawTitle()
{
  WBPoint statusPt = { 50, 50 };
  char buf[255];
  
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "W4: Wiimote Whiteboard With Wii",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "2008 Bjoern Giesler <bjoern@giesler.de>",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Port of a PC app by Johnny Chung Lee",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Kudos to Team Twiizers, svpe and shagkur",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Kudos to the GRRLib Developers",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  statusPt.v += font5_char_high;

  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Instructions:",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;

  snprintf(buf, 255,
	       "Point Wiimote #%d at the screen. Make sure it",
	   wbwiimotenum+1);
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       buf, font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "sees all of the screen, but not too much more.",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "You can use the test screen to see what it sees.",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;  
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Then, calibrate and start drawing!",
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;
  snprintf(buf, 255, "All button-pushing is done on Wiimote #%d.",
	   wiimotenum+1);
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       buf,
	       font5, 0x0000, 0xffff);
  statusPt.v += font5_char_high;  
  statusPt.v += font5_char_high;  
  
  if(WB_GetStatus() == WB_NOTCONNECTED)
    {
      snprintf(buf, 255, "Please connect Wiimote #%d.", wbwiimotenum+1);
      GRRLIB_Print(statusPt.u, statusPt.v,
		   font5_char_width, font5_char_high,
		   buf,
		   font5, 0x0000, 0xffff);
      statusPt.v += font5_char_high;
      statusPt.v += font5_char_high;
    }
  else
    {
      snprintf(buf, 255, "Wiimote #%d connected.", wbwiimotenum+1);
      GRRLIB_Print(statusPt.u, statusPt.v,
		   font5_char_width, font5_char_high,
		   buf,
		   font5, 0x0000, 0xffff);
      statusPt.v += font5_char_high;
      statusPt.v += font5_char_high;
      GRRLIB_Print(statusPt.u, statusPt.v,
		   font5_char_width, font5_char_high,
		   "A: Calibrate & Go, B: Show test Screen",
		   font5, 0x0000, 0xffff);  
      statusPt.v += font5_char_high;
    }
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "HOME: Reboot, +/-: Swap Wiimotes",
	       font5, 0x0000, 0xffff);
}

void drawTest()
{
  WBPoint statusPt = { 100, 100 };

  char buf[255]; int i;

  statusPt.v -= font5_char_high;
  sprintf(buf, "Wiimote #%d camera view:", wbwiimotenum+1);
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       buf, font5, 0x0000, 0xffff);  
  statusPt.v += font5_char_high;

  // show camera screen
  GRRLIB_DrawRectangle(statusPt.u, statusPt.v,
		       128, 128, 0x0000, 1);
  for(i=0; i<WPAD_MAX_IR_DOTS; i++)
    {
      ir_dot_t dot = wpads[wbwiimotenum].ir.dot[i];
      if(dot.x == 0 && dot.y == 0) continue;
      if(WB_IsCloseToStaticDot(dot))
	GRRLIB_DrawRectangle(statusPt.u + wpads[wbwiimotenum].ir.dot[i].x / 8,
			     statusPt.v + wpads[wbwiimotenum].ir.dot[i].y / 6,
			     2, 2, 0xfff0, 1);
      else
	GRRLIB_DrawRectangle(statusPt.u + wpads[wbwiimotenum].ir.dot[i].x / 8,
			     statusPt.v + wpads[wbwiimotenum].ir.dot[i].y / 6,
			     2, 2, 0x0fff, 1);
    }
  for(i=0; i<numCalibPointsDone; i++)
    {
      WBPoint p = normToWiimote(calibPoints2[i]);
      GRRLIB_DrawRectangle(statusPt.u + p.u / 8,
			   statusPt.v + p.v / 6,
			   2, 2, 0x0ff0, 1);
    }
  statusPt.v += 128;
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Static dot", font5, 0xfff0, 0x0000);
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Dynamic dot", font5, 0x0fff, 0x0000);  
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Calibrated dot", font5, 0x0ff0, 0x0000);  
  statusPt.v += font5_char_high;
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "A to return, B to reset static points",
	       font5, 0x0000, 0xffff);  
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "+/- to swap wiimotes",
	       font5, 0x0000, 0xffff);  
}

void drawCalibrating()
{
  WBPoint statusPt = { 100, 100 };
  int i;

  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Calibrating.",
	       font5, 0x0000, 0xffff);  
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Touch the yellow square with your lightpen.",
	       font5, 0x0000, 0xffff);  
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Press A at any time to calibrate again.",
	       font5, 0x0000, 0xffff);  
  statusPt.v += font5_char_high;
  GRRLIB_Print(statusPt.u, statusPt.v,
	       font5_char_width, font5_char_high,
	       "Press B to return to title.",
	       font5, 0x0000, 0xffff);  
  statusPt.v += font5_char_high;

  // show camera screen
  GRRLIB_DrawRectangle(statusPt.u, statusPt.v,
		       128, 128, 0x0000, 1);
  for(i=0; i<WPAD_MAX_IR_DOTS; i++)
    {
      ir_dot_t dot = wpads[wbwiimotenum].ir.dot[i];
      if(dot.x == 0 && dot.y == 0) continue;
      if(WB_IsCloseToStaticDot(dot))
	GRRLIB_DrawRectangle(statusPt.u + wpads[wbwiimotenum].ir.dot[i].x / 8,
			     statusPt.v + wpads[wbwiimotenum].ir.dot[i].y / 6,
			     2, 2, 0xfff0, 1);
      else
	GRRLIB_DrawRectangle(statusPt.u + wpads[wbwiimotenum].ir.dot[i].x / 8,
			     statusPt.v + wpads[wbwiimotenum].ir.dot[i].y / 6,
			     2, 2, 0x0fff, 1);
    }
  for(i=0; i<numCalibPointsDone; i++)
    {
      WBPoint p = normToWiimote(calibPoints2[i]);
      GRRLIB_DrawRectangle(statusPt.u + p.u / 8,
			   statusPt.v + p.v / 6,
			   2, 2, 0x0ff0, 1);
    }
  statusPt.v += 128;

  char buf[255];
  for(i=0; i<numCalibPointsDone; i++)
    {
      sprintf(buf, "%d: %.3f, %.3f",
	      i, calibPoints2[i].u, calibPoints2[i].v);
      GRRLIB_Print(statusPt.u, statusPt.v,
		   font5_char_width, font5_char_high,
		   buf, font5, 0x0000, 0xffff);  
      statusPt.v += font5_char_high;
    }

  // show calibration points
  for(i=0; i<numCalibPoints; i++)
    {
      WBPoint pt = normToScreen(calibPoints1[i]);
      u16 color = 0x0000;
      if(i < numCalibPointsDone)
	color = 0x0ff0;
      else if(i == numCalibPointsDone)
	color = 0xff00;
      GRRLIB_DrawRectangle(pt.u - calibPointRadius, pt.v - calibPointRadius,
			   2*calibPointRadius + 1, 2*calibPointRadius + 1,
			   color, 1);
    }
}

void drawArrow(int x, int y, int width, int height, u16 color, int right)
{
  if(right)
    {
      GRRLIB_Line(x, y, x+width, y+height/2, color);
      GRRLIB_Line(x+width, y+height/2, x, y+height, color);
    }
  else
    {
      GRRLIB_Line(x+width, y, x, y+height/2, color);
      GRRLIB_Line(x, y+height/2, x+width, y+height, color);
    }
}

void drawDrawing()
{
  if(needsfillscreen)
    {
      GRRLIB_FillScreen(0xffff);
      needsfillscreen = 0;
    }

  if(!isPointZero(cursor) && !isPointZero(oldcursor))
    {
      WBPoint pt1 = normToScreen(oldcursor);
      WBPoint pt2 = normToScreen(cursor);
      
      switch(tool)
	{
	case TOOL_LINE:
	  GRRLIB_Line(pt1.u, pt1.v, pt2.u, pt2.v, colors[colorindex]);
	  break;
	case TOOL_SMALLDOT:
	  GRRLIB_DrawRectangle(pt1.u-5, pt1.v-5, 10, 10, colors[colorindex], 1);
	  break;
	case TOOL_LARGEDOT:
	  GRRLIB_DrawRectangle(pt1.u-10, pt1.v-10, 20, 20, colors[colorindex], 1);
	  break;
	}
    }

  // color
  int x = 15;
  int y = 420;
  GRRLIB_DrawRectangle(x, y, 30, 20, 0xffff, 1);
  GRRLIB_DrawRectangle(x, y, 30, 20, 0x0000, 0);
  drawArrow(x, y, 20, 30, 0x0000, 0);
  coldecrP1 = screenToNormXY(x, y);
  coldecrP2 = screenToNormXY(x+20, y+30);
  x+=20;
  GRRLIB_DrawRectangle(x, y, 30, 30, 0xffff, 1);
  GRRLIB_DrawRectangle(x, y, 30, 30, 0x0000, 0);
  GRRLIB_DrawRectangle(x+3, y+3, 24, 24, colors[colorindex], 1);
  x+=30;
  GRRLIB_DrawRectangle(x, y, 30, 20, 0xffff, 1);
  GRRLIB_DrawRectangle(x, y, 30, 20, 0x0000, 0);
  drawArrow(x, y, 20, 30, 0x0000, 1);
  colincrP1 = screenToNormXY(x, y);
  colincrP2 = screenToNormXY(x+20, y+30);
  x+=30;

  // tool type
  GRRLIB_DrawRectangle(x, y, 30, 20, 0xffff, 1);
  GRRLIB_DrawRectangle(x, y, 30, 20, 0x0000, 0);
  drawArrow(x, y, 20, 30, 0x0000, 0);
  tooldecrP1 = screenToNormXY(x, y);
  tooldecrP2 = screenToNormXY(x+20, y+30);
  x+=20;
  GRRLIB_DrawRectangle(x, y, 30, 30, 0xffff, 1);
  GRRLIB_DrawRectangle(x, y, 30, 30, 0x0000, 0);
  switch(tool)
    {
    case TOOL_LINE:
      GRRLIB_Line(x+5, y+5, x+20, y+20, 0x0000);
      break;
    case TOOL_SMALLDOT:
      GRRLIB_DrawRectangle(x+10, y+10, 10, 10, 0x0000, 1);
      break;
    case TOOL_LARGEDOT:
      GRRLIB_DrawRectangle(x+5, y+5, 20, 20, 0x0000, 1);
    } 
  x+=30;
  GRRLIB_DrawRectangle(x, y, 30, 20, 0xffff, 1);
  GRRLIB_DrawRectangle(x, y, 30, 20, 0x0000, 0);
  toolincrP1 = screenToNormXY(x, y);
  toolincrP2 = screenToNormXY(x+20, y+30);
  drawArrow(x, y, 20, 30, 0x0000, 1);
}

int main(int argc, char **argv)
{
  GRRLIB_buffer = (u16*)malloc(screenDim.u * screenDim.v * 2);
  
  VIDEO_Init();
  PAD_Init();

  GRRLIB_InitVideo();
  GRRLIB_Start();
  GRRLIB_Render();
  
  VIDEO_WaitVSync();
  
  WPAD_Init();
  WPAD_SetDataFormat(WPAD_CHAN_0, WPAD_FMT_CORE_ACC_IR);
  WPAD_SetDataFormat(WPAD_CHAN_1, WPAD_FMT_CORE_ACC_IR);

  WB_Init(1);
  
  wpads = WB_GetWPADData();

  colorindex = 0;
  tool = TOOL_LINE;
  
  while(1)
    {
      if(state != STATE_DRAWING)
	{
	  GRRLIB_FillScreen(0xffff);
	}

      PAD_ScanPads();
      WPAD_ScanPads();
      if(PAD_ButtonsDown(0) & PAD_BUTTON_START ||
	 buttonPushed(WPAD_BUTTON_HOME))
	reboot();

      WB_Read();

      WBStatus status = WB_GetStatus();
      if(state == STATE_TITLE)
	{
	  drawTitle();
	  if(buttonPushed(WPAD_BUTTON_A) &&
	     status != WB_NOTCONNECTED)
	    {
	      WB_Init(wbwiimotenum);
	      numCalibPointsDone = 0;
	      state = STATE_CALIBRATING;
	    }
	  else if(buttonPushed(WPAD_BUTTON_B) &&
		  status != WB_NOTCONNECTED)
	    {
	      WB_Init(wbwiimotenum);
	      state = STATE_TEST;
	    }
	  else if(buttonPushed(WPAD_BUTTON_PLUS) ||
		  buttonPushed(WPAD_BUTTON_MINUS))
	    {
	      wiimotenum = !wiimotenum; wbwiimotenum = !wbwiimotenum;
	    }
	}
      else if(state == STATE_TEST)
	{
	  drawTest();
	  if(buttonPushed(WPAD_BUTTON_A) &&
	     status != WB_NOTCONNECTED)
	    {
	      WB_Init(wbwiimotenum);
	      state = STATE_TITLE;
	    }
	  else if(buttonPushed(WPAD_BUTTON_B) &&
	     status != WB_NOTCONNECTED)
	    {
	      WB_Init(wbwiimotenum);
	    }
	  else if(buttonPushed(WPAD_BUTTON_PLUS) ||
		  buttonPushed(WPAD_BUTTON_MINUS))
	    {
	      wiimotenum = !wiimotenum; wbwiimotenum = !wbwiimotenum;
	    }
	}
      else if(state == STATE_CALIBRATING)
	{
	  if(buttonPushed(WPAD_BUTTON_A) &&
	     status != WB_NOTCONNECTED)
	    {
	      WB_Init(wbwiimotenum);
	      numCalibPointsDone = 0;
	      state = STATE_CALIBRATING;
	    }
	  else if(buttonPushed(WPAD_BUTTON_B))
	    {
	      WB_Init(wbwiimotenum);
	      numCalibPointsDone = 0;
	      state = STATE_TITLE;
	    }

	  drawCalibrating();

	  ir_dot_t nsdots[WPAD_MAX_IR_DOTS]; int numnsdots;
	  WB_GetNonStaticDots(nsdots, &numnsdots);

	  if(numnsdots == 0 && held.u != 0 && held.v != 0)
	    {
	      WBPoint pt = wiimoteToNorm(held);
	      if(numCalibPointsDone == 0 ||
		 distance(pt, calibPoints2[numCalibPointsDone-1]) > 0.05)
		{
		  calibPoints2[numCalibPointsDone] = pt;
		  numCalibPointsDone++;
		  if(numCalibPointsDone >= numCalibPoints)
		    {
		      WB_WarpSetSource(calibPoints2[0], calibPoints2[1],
				       calibPoints2[2], calibPoints2[3]);
		      WB_WarpSetDest(calibPoints1[0], calibPoints1[1],
				     calibPoints1[2], calibPoints1[3]);
		      needsfillscreen = 1;
		      state = STATE_DRAWING;
		    }
		}
	      else sprintf(msg, "Too far away!");
	      held = zero;
	    }
	  else if(numnsdots == 1 && held.u == 0 && held.v == 0)
	    {
	      held.u = nsdots[0].x; held.v = nsdots[0].y;
	    }
	}
      else // state == STATE_DRAWING
	{
	  if(buttonPushed(WPAD_BUTTON_A) &&
	     status != WB_NOTCONNECTED)
	    {
	      WB_Init(wbwiimotenum);
	      numCalibPointsDone = 0;
	      state = STATE_TITLE;
	    }

	  ir_dot_t nsdots[WPAD_MAX_IR_DOTS]; int numnsdots;
	  WB_GetNonStaticDots(nsdots, &numnsdots);

	  if(numnsdots == 1) // drawing and one point visible
	    {
	      WBPoint pt = {nsdots[0].x, nsdots[0].y};
	      pt = wiimoteToNorm(pt);
	      pt = WB_Warp(pt);

	      if(isPointZero(cursor))
		startcursor = pt;
	      oldcursor = cursor;
	      cursor = pt;
	    }
	  else
	    {
	      if(!isPointZero(startcursor) && !isPointZero(oldcursor))
		wbclick(startcursor, oldcursor);
	      oldcursor = cursor = zero;
	    }

	  drawDrawing();
	}
      GRRLIB_Render();
      VIDEO_WaitVSync();
    }

  return 0;
}


extern void SYS_PreMain();
int __crtmain() {
  SYS_PreMain();
  return main(0, NULL);
}
