/* This is -*- C -*-)

   wbwiimote.c   
   \author: Bjoern Giesler <bjoern@giesler.de>
   
   
   
   $Author: giesler $
   $Locker$
   $Revision$
   $Date: 2002-08-19 10:41:28 +0200 (Mon, 19 Aug 2002) $
 */

/* system includes */
#include <wiiuse/wpad.h>
#include <math.h>
#include <unistd.h>

/* my includes */
#include "wbwiimote.h"

static int wiimotenum = -1;
static WPADData wpads[4];
static ir_dot_t staticdots[WPAD_MAX_IR_DOTS];
static int numstaticdots;

static WBPoint src[4], dst[4];
static float srcMat[16], dstMat[16], warpMat[16];
static int dirty;

/* Forward declaration of private functions
   (Implementation is below) */
static void WBP_WarpCompute();
static void WBP_MultMats(float* srcMat, float* dstMat, float* resMat);
static void WBP_SquareToQuad(WBPoint p0, WBPoint p1, WBPoint p2, WBPoint p3,
			     float* mat);
static void WBP_QuadToSquare(WBPoint p0, WBPoint p1, WBPoint p2, WBPoint p3,
			     float* mat);

void WB_Init(int num)
{
  wiimotenum = num;

  WB_Read();
  int i;
  numstaticdots = 0;
  for(i=0; i<WPAD_MAX_IR_DOTS; i++)
    {
      ir_dot_t dot = wpads[wiimotenum].ir.dot[i];
      if(dot.x != 0 || dot.y != 0)
	{
	  staticdots[i] = dot;
	  numstaticdots++;
	}
    }
  WB_WarpSetIdentity();
}

void WB_Read()
{
  memset(wpads, 0, 4*sizeof(WPADData));
  WPAD_Read(0, &(wpads[0]));
  WPAD_Read(1, &(wpads[1]));
}

WPADData* WB_GetWPADData() { return wpads; }
int WB_GetWiimoteNum() { return wiimotenum; }

WBStatus WB_GetStatus()
{
  if(wiimotenum == -1) return WB_NOTINITED;
  
  if(wpads[wiimotenum].accel.x == 0 &&
     wpads[wiimotenum].accel.y == 0 &&
     wpads[wiimotenum].accel.z == 0)
    return WB_NOTCONNECTED;

  ir_dot_t nsdots[WPAD_MAX_IR_DOTS];
  int numnsdots;
  WB_GetNonStaticDots(nsdots, &numnsdots);
  switch(numnsdots)
    {
    case 0:
      return WB_SEESNONE;
    case 1:
      return WB_SEESONE;
    default:
      return WB_SEESMORE;
    }
}

int WB_IsCloseToStaticDot(ir_dot_t dot)
{
  int i;
  
  for(i=0; i<WPAD_MAX_IR_DOTS; i++)
    {
      if(staticdots[i].x == 0 && staticdots[i].y == 0) continue;
      int distx = staticdots[i].x - dot.x;
      if(distx < 0) distx = -distx;
      int disty = staticdots[i].y - dot.y;
      if(disty < 0) disty = -disty;
      if(distx+disty < 10)
	{
	  return 1;
	}
    }
  return 0;
}

void WB_GetNonStaticDots(ir_dot_t nsdots[WPAD_MAX_IR_DOTS], int* numnsdots)
{
  *numnsdots = 0;

  int i, j;
  for(j=0; j<WPAD_MAX_IR_DOTS; j++)
    {
      int found = 0;
      if(wpads[wiimotenum].ir.dot[j].x == 0 &&
	 wpads[wiimotenum].ir.dot[j].y == 0)
	continue;

      for(i=0; i<numstaticdots; i++)
	{
	  if(staticdots[i].x == 0 && staticdots[i].y == 0) continue;
	  int distx = staticdots[i].x - wpads[wiimotenum].ir.dot[j].x;
	  if(distx < 0) distx = -distx;
	  int disty = staticdots[i].y - wpads[wiimotenum].ir.dot[j].y;
	  if(disty < 0) disty = -disty;
	  if(distx+disty < 10)
	    {
	      found = 1;
	      break;
	    }
	}
      if(!found)
	{
	  nsdots[*numnsdots] = wpads[wiimotenum].ir.dot[j];
	  (*numnsdots)++;
	}
    }
}

void WB_WarpSetIdentity()
{
  WBPoint p0 = {0, 0}, p1 = {1, 0}, p2 = {0, 1}, p3 = {1, 1};
  WB_WarpSetSource(p0, p1, p2, p3);
  WB_WarpSetDest(p0, p1, p2, p3);
  WBP_WarpCompute();
}

void WB_WarpSetSource(WBPoint p0, WBPoint p1, WBPoint p2, WBPoint p3)
{
  src[0].u = p0.u; src[0].v = p0.v; src[1].u = p1.u; src[1].v = p1.v; 
  src[2].u = p2.u; src[2].v = p2.v; src[3].u = p3.u; src[3].v = p3.v;
  dirty = 1;
}

void WB_WarpSetDest(WBPoint p0, WBPoint p1, WBPoint p2, WBPoint p3)
{
  dst[0].u = p0.u; dst[0].v = p0.v; dst[1].u = p1.u; dst[1].v = p1.v; 
  dst[2].u = p2.u; dst[2].v = p2.v; dst[3].u = p3.u; dst[3].v = p3.v;
  dirty = 1;
}

WBPoint WB_Warp(WBPoint p)
{
  if(dirty) WBP_WarpCompute();
  float result[4];
  WBPoint retval;

  result[0] = (float)(p.u * warpMat[0] + p.v*warpMat[4] + warpMat[12]);
  result[1] = (float)(p.u * warpMat[1] + p.v*warpMat[5] + warpMat[13]);
  result[2] = (float)(p.u * warpMat[2] + p.v*warpMat[6] + warpMat[14]);
  result[3] = (float)(p.u * warpMat[3] + p.v*warpMat[7] + warpMat[15]);

  retval.u = result[0]/result[3];
  retval.v = result[1]/result[3];

  return retval;
}

/* Private functions */

void WBP_WarpCompute()
{
  WBP_QuadToSquare(src[0], src[1], src[2], src[3], srcMat);
  WBP_SquareToQuad(dst[0], dst[1], dst[2], dst[3], dstMat);
  WBP_MultMats(srcMat, dstMat, warpMat);
  dirty = 0;
}

void WBP_MultMats(float* srcMat, float* dstMat, float* resMat)
{
  int r, c, ri;
  for(r=0; r<4; r++) 
    {
      ri = r * 4;
      for(c=0; c<4; c++) 
	{
	  resMat[ri+c] = (srcMat[ri  ] * dstMat[c   ] +
			  srcMat[ri+1] * dstMat[c+ 4] +
			  srcMat[ri+2] * dstMat[c+ 8] +
			  srcMat[ri+3] * dstMat[c+12]);
	}
    }
}

void WBP_SquareToQuad(WBPoint p0, WBPoint p1, WBPoint p2, WBPoint p3,
		      float* mat)
{
  float dx1 = p1.u - p2.u, 	dy1 = p1.v - p2.v;
  float dx2 = p3.u - p2.u, 	dy2 = p3.v - p2.v;
  float sx = p0.u - p1.u + p2.u - p3.u;
  float sy = p0.v - p1.v + p2.v - p3.v;
  float g = (sx * dy2 - dx2 * sy) / (dx1 * dy2 - dx2 * dy1);
  float h = (dx1 * sy - sx * dy1) / (dx1 * dy2 - dx2 * dy1);
  float a = p1.u - p0.u + g * p1.u;
  float b = p3.u - p0.u + h * p3.u;
  float c = p0.u;
  float d = p1.v - p0.v + g * p1.v;
  float e = p3.v - p0.v + h * p3.v;
  float f = p0.v;

  mat[ 0] = a;	mat[ 1] = d;	mat[ 2] = 0;	mat[ 3] = g;
  mat[ 4] = b;	mat[ 5] = e;	mat[ 6] = 0;	mat[ 7] = h;
  mat[ 8] = 0;	mat[ 9] = 0;	mat[10] = 1;	mat[11] = 0;
  mat[12] = c;	mat[13] = f;	mat[14] = 0;	mat[15] = 1;
}

void WBP_QuadToSquare(WBPoint p0, WBPoint p1, WBPoint p2, WBPoint p3,
		      float* mat)
{
  WBP_SquareToQuad(p0, p1, p2, p3, mat);

  float a = mat[ 0], d = mat[ 1], /* ignore */ g = mat[ 3];
  float b = mat[ 4], e = mat[ 5], /* 3rd col*/ h = mat[ 7];
  /* ignore 3rd row */
  float c = mat[12], f = mat[13];

  float A = e - f * h;
  float B = c * h - b;
  float C = b * f - c * e;
  float D = f * g - d;
  float E =     a - c * g;
  float F = c * d - a * f;
  float G = d * h - e * g;
  float H = b * g - a * h;
  float I = a * e - b * d;

  // Probably unnecessary since 'I' is also scaled by the determinant,
  //   and 'I' scales the homogeneous coordinate, which, in turn,
  //   scales the X,Y coordinates.
  // Determinant  =   a * (e - f * h) + b * (f * g - d) + c * (d * h - e * g);

  float idet = 1.0f / (a * A           + b * D           + c * G);

  mat[ 0] = A * idet; mat[ 1] = D * idet; mat[ 2] = 0; mat[ 3] = G * idet;
  mat[ 4] = B * idet; mat[ 5] = E * idet; mat[ 6] = 0; mat[ 7] = H * idet;
  mat[ 8] = 0       ; mat[ 9] = 0       ; mat[10] = 1; mat[11] = 0       ;
  mat[12] = C * idet; mat[13] = F * idet; mat[14] = 0; mat[15] = I * idet;
}

